from abc import ABC, abstractmethod
from typing import Callable, Union

class Option[T](ABC):
    """Abstract base class for NoneOption and SomeOption."""

    @abstractmethod
    def get(self) -> T:
        """Returns the option's value."""
        pass

    @abstractmethod
    def is_empty(self) -> bool:
        """Check if the option is empty."""
        pass
    
    @abstractmethod
    def non_empty(self) -> bool:
        """Check if the option is non-empty."""
        pass

    @abstractmethod
    def get_or_else(self, default: Union[T, Callable[[], T]]) -> T:
        """
        Returns the option's value if non-empty, otherwise returns the default.
        Supports both eager and lazy evaluation.
        """
        pass

    @abstractmethod
    def or_else(self, alternative: Callable[[], 'Option[T]']) -> 'Option[T]':
        """
        Returns this option if non-empty, otherwise returns the alternative.
        Lazy evaluation of alternative.
        """
        pass

    @abstractmethod
    def map[U](self, f: Callable[[T], U]) -> 'Option[U]':
        """Applies the function to the option's value if non-empty."""
        pass

    @abstractmethod
    def flat_map[U](self, f: Callable[[T], 'Option[U]']) -> 'Option[U]':
        """
        Applies the function that returns an Option to the option's value.
        """
        pass

    @abstractmethod
    def fold[U](self, if_empty: Callable[[], U], f: Callable[[T], U]) -> U:
        """
        If the option is empty, applies the 'if_empty' function and returns its result.
        If the option is non-empty, applies the function 'f' to the option's value.
        """
        pass

    @abstractmethod
    def filter(self, p: Callable[[T], bool]) -> 'Option[T]':
        """
        Returns the option if non-empty and the predicate returns True.
        """
        pass

    @abstractmethod
    def exists(self, p: Callable[[T], bool]) -> bool:
        """
        Returns True if non-empty and the predicate returns True.
        """
        pass

    @abstractmethod
    def contains(self, elem: T) -> bool:
        """
        Checks if the option contains the given element.
        """
        pass

    @abstractmethod
    def to_list(self) -> list[T]:
        """
        Converts the option to a list.
        """
        pass



from typing import Callable, Union
from .exceptions import NoSuchElementException
from .option import Option

class NoneOption(Option[None]):
    def __repr__(self) -> str:
        return "none"

    def get(self) -> None:
        raise NoSuchElementException("Called get on NoneOption")

    def is_empty(self) -> bool:
        return True

    def non_empty(self) -> bool:
        return False

    def get_or_else[T](self, default: Union[T, Callable[[], T]]) -> T:
        return default() if callable(default) else default

    def or_else[T](self, alternative: Callable[[], Option[T]]) -> Option[T]:
        return alternative()

    def map[T, U](self, f: Callable[[T], U]) -> 'NoneOption':
        return none

    def flat_map[T](self, f: Callable[[T], None]) -> 'NoneOption':
        return none

    def fold[T, U](self, if_empty: Callable[[], U], f: Callable[[T], U]) -> U:
        return if_empty()

    def filter[T](self, p: Callable[[T], bool]) -> 'NoneOption':
        return none

    def exists[T](self, p: Callable[[T], bool]) -> bool:
        return False

    def contains[T](self, elem: T) -> bool:
        return False

    def to_list[T](self) -> list[T]:
        return []

none: NoneOption = NoneOption()
"""Singleton NoneOption instance"""



from typing import Callable, Union
from .none_option import none
from .option import Option

class SomeOption[T](Option[T]):

    def __init__(self, value: T) -> Option[T]:
        self._value = value

    def __repr__(self) -> str:
        return f"some({self._value})"

    def get(self) -> T:
        return self._value

    def is_empty(self) -> bool:
        return False
    
    def non_empty(self) -> bool:
        return True

    def get_or_else(self, default: Union[T, Callable[[], T]]) -> T:
        return self._value

    def or_else(self, alternative: Callable[[], Option[T]]) -> Option[T]:
        return self

    def map[U](self, f: Callable[[T], U]) -> 'SomeOption[U]':
        return some(f(self._value))

    def flat_map[U](self, f: Callable[[T], Option[U]]) -> Option[U]:
        return f(self._value)
    
    def fold[U](self, if_empty: Callable[[], U], f: Callable[[T], U]) -> U:
        return f(self._value)

    def filter(self, p: Callable[[T], bool]) -> Option[T]:
        return self if p(self._value) else none

    def exists(self, p: Callable[[T], bool]) -> bool:
        return p(self._value)

    def contains(self, elem: T) -> bool:
        return self._value == elem

    def to_list(self) -> list[T]:
        return [self._value]

# Factory function for creating SomeOption objects
def some[T](value: T) -> Option[T]:
    """Create a SomeOption with the given value."""
    return SomeOption(value)
